
{

 Definition of all public datatypes of Ultra Pascal

 created by MXM-TriX
 alexander.grutsch.htlr@schulen.vol.at

 under the terms of MXM Projects
 http://start.at/mxm

}

unit upas_def;

interface

uses strings,upas_sys;

type

      str80	= string[80];
      str10	= string[10];
      { var_types       Definition of all possible Pascal datatypes }
      var_types = (_byte,_word,_longint,_pointer,_real,_string,_record,
                   _array,_array_record,_no_type,_char, _dword, _shortint,
		   _whitespaces, _sprites, _external);

      { keywords        Definition of all Ultra Pascal keywords }
      keywords  = (_program,_uses,_begin,_end,_var,_procedure,_function,
                   _if,_while,_for,_repeat, __record,_export,
                   _math,_const,_def_lib,_until,_import, _forward,
                   _interrupt,_asm,_assembly,_synonym,_unit,
                   _librarydefinition,_interface,_implementation,_own_call,
		   _to, _then, _of, _do, _continue, _break, _exit, _halt,
		   _else, _type, _case, _try);

      { Variables are stored in a list. [Last var].next_var = nil; }
      ppascal_var  = ^pascal_var;
      pascal_var   = record
       name                : string[40];
       rechte              : boolean;      { Used for const }
       size                : word;
       stackpos            : integer;
       var_set		   : boolean;
       var_t_size	   : word;         { real datatypesize }
       variabelntyp        : var_types;
       arraytyp            : var_types;
       record_first_var    : ppascal_var;  { First record variable }
       next_var            : ppascal_var;  { Next variable address }
       previous_var	   : ppascal_var;
       descriptor_string   : string[80];
       const_string        : string[50];
       prev_c		   : char;
       needed_register     : string[5];    { Used for librarycalling }
       call_level	   : word;	   { Used for right calling of the
        				     parameters from a library }
      end;

      calc_registers    = record
        name: string[4];
        in_use: boolean;
      end;

      ppascal_procedure = ^pascal_procedure;
      pascal_procedure = record
        name           : string[40];
        synonym        : string[80];
        funktion       : boolean;
        fkt_result     : var_types;
        first_var      : ppascal_var;
        last_var       : ppascal_var;
        fkt_var        : ppascal_var;
        next_procedure : ppascal_procedure;
        result_in      : string[6];
        used           : boolean; { Used, to truncate unused procedure }
        code_created   : boolean; { Used, for FORWARD definitions }
        white_spaces   : boolean; { Used for calling with unlimited number
                                    of parameters }
        asm_func       : boolean;
      end;

      pdefinitions  = ^tdefinitions;
      definitions   = record
        name: string[40];
        defined: boolean;
      end;
      tdefinitions  = array[0..99]of definitions;

      pinclude_libs = ^tinclude_libs;
      include_libs  = record
        lib_name    : string[12];
        unit_bool   : boolean;
        imported    : boolean;
      end;
      tinclude_libs = array[0..9] of include_libs;

      pascalkeyword = record
        name        : string[20];
        keyreturn   : keywords;
        lastword    : boolean;
        isoneword   : boolean;
      end;

      var_type_definition_align = record
       is_float                 : boolean;
       lo_size, hi_size         : longint;
       name                     : string;
       savestr                  : char;
       size                     : word;
       v_type                   : var_types;
       write_how                : string[8];
       w_str                    : boolean;
       signed			: boolean;
      end;

      operators                 = record
       name                   : string[5];
       use_in_float           : boolean;
       float_equal            : string[10];
       non_float_equal        : string[8];
       u_non_float_equal      : string[8];
       branch_code_signed     : string[5];
       branch_code_unsigned   : string[5];
       {
         REG_*                Used for registers, which should be used, ''
         when all registers are allowed
       }
       reg_1                  : string[2];
       reg_2                  : string[2];
      end;
      defination_operands = record
                name: char;
                tomodify: word;
                op_pos: integer;
          end;
      numeric_systems = record
        identification : char;
        allowed_chars  : string[20];
        base           : longint;
      end;

const   pascal_keywords: array[0..39] of pascalkeyword =
       ((name: 'program '; keyreturn: _program; lastword: false; isoneword: true),
        (name: 'librarydefinition'; keyreturn: _librarydefinition; lastword: true; isoneword: true),
        (name: 'library'; keyreturn: _def_lib; lastword: false; isoneword: true),
        (name: 'unit '; keyreturn: _unit; lastword: false; isoneword: true),
        (name: 'uses '; keyreturn: _uses; lastword: false; isoneword: true),
        (name: 'interface'; keyreturn: _interface; lastword: true; isoneword: true),
        (name: 'implementation'; keyreturn: _implementation; lastword: true; isoneword: true),
        (name: 'begin'; keyreturn: _begin; lastword: true; isoneword: true),
        (name: 'synonym '; keyreturn: _synonym; lastword: false; isoneword: true),
        (name: 'end'; keyreturn: _end; lastword: true; isoneword: true),
        (name: 'var'; keyreturn: _var; lastword: true; isoneword: true),
        (name: 'type'; keyreturn: _type; lastword: true; isoneword: true),
        (name: 'procedure ';keyreturn: _procedure; lastword: false; isoneword: true),
        (name: 'function ';keyreturn: _function; lastword: false; isoneword: true),
        (name: 'if ';keyreturn: _if; lastword: false; isoneword: true),
        (name: 'for '; keyreturn: _for; lastword: false; isoneword: true),
        (name: 'while '; keyreturn: _while; lastword: false; isoneword: true),
        (name: 'repeat'; keyreturn: _repeat; lastword: true; isoneword: true),
        (name: 'record'; keyreturn: __record; lastword: true; isoneword: true),
        (name: 'export'; keyreturn: _export; lastword: true; isoneword: true),
        (name: ':='; keyreturn: _math; lastword: false; isoneword: false),
        (name: 'until '; keyreturn: _until; lastword: false; isoneword: true),
        (name: 'const'; keyreturn: _const; lastword: true; isoneword: true),
        (name: 'forward'; keyreturn: _forward; lastword: false; isoneword: true),
        (name: 'interrupt'; keyreturn: _interrupt; lastword: false; isoneword: true),
        (name: 'asm'; keyreturn: _asm; lastword: true; isoneword: true),
        (name: 'assembler'; keyreturn: _assembly; lastword: false; isoneword: false),
	(name: 'do'; keyreturn: _do; lastword: true; isoneword: true),
	(name: 'then'; keyreturn: _then; lastword: true; isoneword: true),
	(name: 'break'; keyreturn: _break; lastword: false; isoneword: true),
	(name: 'continue'; keyreturn: _continue; lastword: false; isoneword: true),
	(name: 'exit'; keyreturn: _exit; lastword: false; isoneword: true),
	(name: 'halt'; keyreturn: _halt; lastword: false; isoneword: true),
	(name: 'else'; keyreturn: _else; lastword: true; isoneword: true),
        (name: 'to'; keyreturn: _to; lastword : false; isoneword: true),
        (name: 'then'; keyreturn: _then; lastword: true; isoneword: true),
        (name: 'of'; keyreturn: _of; lastword: false; isoneword: true),
        (name: 'do'; keyreturn: _do; lastword: true; isoneword: true),
        (name: 'case'; keyreturn: _case; lastword: true; isoneword: true),
        (name: 'try'; keyreturn: _try; lastword: true; isoneword: true));

     { Alle Operatoren, nach Prioritten geordnet }
     pascal_operands: array[0..22] of operators =
((name: '*'; use_in_float: true; float_equal: 'ffpmul2'; non_float_equal: 'muls'; u_non_float_equal: 'mulu';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: 'd6'; reg_2: 'd7'),
 (name: '/'; use_in_float: true; float_equal: 'ffpdiv'; non_float_equal: 'divs'; u_non_float_equal: 'divu';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: 'd6'; reg_2: 'd7'),
 (name: ' mod '; use_in_float: false; float_equal: ''; non_float_equal: ''; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' shr '; use_in_float: false; float_equal: ''; non_float_equal: 'lsr'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' shl '; use_in_float: false; float_equal: ''; non_float_equal: 'lsl'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' div '; use_in_float: false; float_equal: ''; non_float_equal: 'divs'; u_non_float_equal: 'divu';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' rol '; use_in_float: false; float_equal: ''; non_float_equal: 'rol'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' ror '; use_in_float: false; float_equal: ''; non_float_equal: 'ror'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: '>>'; use_in_float: false; float_equal: ''; non_float_equal: 'lsr'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: '<<'; use_in_float: false; float_equal: ''; non_float_equal: 'lsl'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' and '; use_in_float: false; float_equal: ''; non_float_equal: 'and'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' or '; use_in_float: false; float_equal: ''; non_float_equal: 'or'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: ' xor '; use_in_float: false; float_equal: ''; non_float_equal: 'eor'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: '&&'; use_in_float: false; float_equal: ''; non_float_equal: 'and'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: '||'; use_in_float: false; float_equal: ''; non_float_equal: 'or'; u_non_float_equal : '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: ''; reg_2: ''),
 (name: '+'; use_in_float: true; float_equal: 'ffpadd'; non_float_equal: 'add'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: 'd6'; reg_2: 'd7'),
 (name: '-'; use_in_float: true; float_equal: 'ffpsub'; non_float_equal: 'sub'; u_non_float_equal: '';
  branch_code_signed: ''; branch_code_unsigned: ''; reg_1: 'd6'; reg_2: 'd7'),
 (name: '>='; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'bge'; branch_code_unsigned: 'bcc'; reg_1: 'd6'; reg_2: 'd7'),
 (name: '<='; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'ble'; branch_code_unsigned: 'bls'; reg_1: 'd6'; reg_2: 'd7'),
 (name: '<>'; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'bne'; branch_code_unsigned: 'bne'; reg_1: 'd6'; reg_2: 'd7'),
 (name: '='; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'beq'; branch_code_unsigned: 'beq'; reg_1: 'd6'; reg_2: 'd7'),
 (name: '>'; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'bgt'; branch_code_unsigned: 'bhi'; reg_1: 'd6'; reg_2: 'd7'),
 (name: '<'; use_in_float: true; float_equal: 'ffpcmp'; non_float_equal: 'cmp'; u_non_float_equal: '';
  branch_code_signed: 'blt'; branch_code_unsigned: 'bcs'; reg_1: 'd6'; reg_2: 'd7'));

    prefix_or_suffix : array[0..2]of defination_operands =
              ((name: '@'; tomodify: 0; op_pos: 1),
               (name: '-'; tomodify: 1; op_pos: 1),
               (name: '^'; tomodify: 2; op_pos: -1));

    rackets       : string[4] = '([)]';


	NULL: BYTE  = $00;
	c_byte_value		= $1;
        c_word_value		= $2;
        c_long_value		= $4;
        c_pointer_value 	= $8;
        c_real_value		= $10;
        c_string_value 	 	= $20;
        c_record_value		= $40;
        c_array_value		= $80;
        c_array_record_value	= $100;
        c_no_type_value		= $200;
        c_char_value		= $400;
        c_dword_value		= $800;
        c_shortint_value	= $1000;
        c_whitespaces_value     = $2000;
        c_sprites_value         = $4000;
        c_external_value        = $8000;

        pascal_numeric_systems: array[0..1] of numeric_systems =
         ((identification: '$'; allowed_chars: '0123456789abcdef'; base: 16),
         (identification: '%'; allowed_chars: '01'; base: 2));

      unit_dir		: string[9] = 'UNITS\';
      sprite_dir	: string[9] = 'SPRITES\';
      table_dir		: string[9] = 'UP-SYS\';
        { -->> Labels >>-- }
        normal_label = 0;
        way_2_label = 1;

{ Queuezuteilung }
        begin_queue		= 6;
        con_queue		= 5;
        stp_queue		= 4;
        end_queue		= 3;
        preprozessor_queue      = 2;
        keyword_queue           = 1;
        label_queue             = 0;

var unit_compile : boolean;
    unit_name	 : string;
    linenr	 : word;
    infile, outfile             : text;
    sourcefile, outputfile  : string;
    predirectory            : string;
    infile_string: PChar;
    infile_str_pos: word;
    lib_def            : boolean;
    variable_data_types     : array[0..12] of var_type_definition_align;
    anz_variable_data_types : word;

{ -->> Variable tree -->> }
        global_first_variable   : ppascal_var;
        global_last_variable    : ppascal_var;

        local_first_variable    : ppascal_var;
        local_last_variable     : ppascal_var;

{ -->> TYPE operator >>-- }
        type_first_variable     : ppascal_var;
        type_last_variable      : ppascal_var;

{ -->> Procedure tree >>-- }
        first_procedure         : ppascal_procedure;
        last_procedure          : ppascal_procedure;
        current_procedure       : ppascal_procedure;

{ Container for Fargo I / Fargo II / 92 + support }
        fargo_string_stream     : pstringstream;
        fargo_version_stream	: pfargo_system;
        fargo_version_str	: string[10];
        { String containers }
        unit_container		: pstringstream;
        form_container_1	: pstringstream;
        form_container_2	: pstringstream;
        form_container_3	: pstringstream;
	temp_container		: pstringstream;

        { File container }
        file_container		: pupasfile;
        standard_string_size    : word;
        apv_counter		: byte;
        used_white_spaces 	: boolean;
        procedure_defined       : boolean;
        procedure_exit_str      : string;
        function_defined        : boolean;
        function_name           : string;
        total_line_counter      : word;
{ Von Preprozessor benutzt, um zu IFDEF Konstruktionen zu ermglichen }
        defined_names           : pdefinitions;
        read_infile             : boolean;
        program_stack_size      : word;
        last_variable_size      : word;
        last_infile_str : string;

    { -->> INFILE operations >>--}
    comments: integer;
    string_read: boolean;
    c_rackets: word;
    end_char: char;
    check_of: boolean;

procedure up_con_write ( s: string );
function conv_to_str ( x: longint ): string;

 procedure up_write ( s: PChar );
 procedure up_gotoxy ( x, y: word );
 procedure up_textattr ( x: word );
 procedure up_clearline;
 procedure up_clrscr;

procedure error (errornum: word);
procedure error_txt(txt: string);
procedure warning (warnnr: word);

function fileexist ( fn: string ): boolean;
function filenewer ( fn1, fn2: string ): boolean;
procedure settime ( fn1, fn2: string );

procedure output (s: string);
procedure output_chars ( s: string );
procedure output_char ( c: char );
procedure output_commentbox(s: string);

procedure call_library ( libname, libfunc: string );

function data_type_index (tname: string): integer;
function build_var_type ( var_str: string; var last_var, first_var: ppascal_var; record_call, rights,
                          parameterbuild: boolean ): longint;
procedure create_new_var (var last_var: ppascal_var);
procedure create_new_procedure;
procedure controll_name (s: string);
function get_next_line: string;

procedure increase_all_variables (size: word; which_var: ppascal_var; inc_last: boolean);
procedure delete_all_variables ( var firstvar, lastvar: ppascal_var; clearall: boolean);
function is_token (x: string; lastwordtest, totallytest: boolean): boolean;
function get_var_size ( s: string ): word;

implementation

{$IFDEF DPMI}
 uses dos,
{$ELSE}
{$IFDEF WINDOWS}
 uses windos,
{$ELSE}
 uses crt,dos,
{$ENDIF}
{$ENDIF}
 upas_uni,upas_opt,upas_str,upas_mat,upas_tk;

function conv_to_str ( x: longint ): string;
var s: string;
begin
 str(x, s);
 conv_to_str := s;
end;

{
 ****************************************************************************
                          Fehler / Warnungsausgabe
 ****************************************************************************
}

const errors: array[0..25] of string[55] =
      ('Unknown variable',
       'Illegal beginning of the program',
       'File ends wrong',
       'Wrong "$ELSE" in preprocessor',
       'Internal error',
       'Illegal racket',
       'Name too long',
       'Illegal character in name',
       'Unknown datatype',
       'Wrong arraydefinition',
       'Illegal character in numeric system',
       'Left side is a constant',
       'Invalid arrayindexes',
       'Invalid new stacksize',
       'Wrong variabledefinition',
       'Unknown datatype',
       'Procedure already defined',
       'Wrong definition of a string',
       'Invalid stringsize',
       'VAR isn''t allowed here',
       'Index is necessary',
       'No procedure defined',
       'Unknown recordvariable',
       '"interface" or "librarydefinition" is here necessary',
       'In a librarydefinition this isn''t allowed',
       'Wrong operation with real'
       );

procedure error (errornum: word);
begin
up_textattr($B);
up_gotoxy ( 1, 18 );
up_clearline;
up_con_write ('Error '+conv_to_str ( errornum)+' in line number '+conv_to_str(linenr));
up_gotoxy (1, 20);
up_clearline;
up_con_write (errors[errornum-1]);
up_textattr (7);
up_gotoxy (1, 19);
up_clearline;
up_con_write (last_infile_str);

close(infile);
if not unit_compile then close(outfile);
halt;
end;

procedure error_txt(txt: string);
begin
up_textattr ($B);
up_gotoxy (1,18);
up_con_write ( txt );
up_gotoxy (1, 19);
up_clearline;
up_con_write (last_infile_str);
{$I-}
close (infile);
{$I+}
if ioresult<>0 then;
if not unit_compile then begin
 {$I-}
 close(outfile);
 if ioresult<>0 then;
 {$I+}
end;
halt;
end;

const warnings: array[0..3] of string[60] =
      ('Too much "}" are ignored',
       'Too much ")" are ignored',
       '"interface" / "implementation" in a program is ignored',
       'Relocation is no calculation -> ignored');

procedure warning (warnnr: word);
begin
up_textattr($B);
up_gotoxy(1,20);
up_con_write('Warning '+conv_to_str(warnnr)+' in line number '+conv_to_str(linenr));
up_gotoxy(1, 21);
up_con_write(warnings [warnnr-1]);
up_textattr($7);
up_gotoxy(1,10);
end;

var up_output: PChar;

procedure up_con_write ( s: string );
begin
 if up_output = nil then
  getmem ( up_output, $200 );
 StrPCopy ( up_output, s );
 up_write ( up_output );
end;

{ -->> Import of the library >>-- }
{$IFDEF DPMI}

 procedure up_write ( s: PChar ); 	external 'UPDPMIWR' index 1;
 procedure up_gotoxy ( x, y: word ); 	external 'UPDPMIWR' index 2;
 procedure up_textattr ( x: word ); 	external 'UPDPMIWR' index 3;
 procedure up_clearline; 		external 'UPDPMIWR' index 4;
 procedure up_clrscr; 		  	external 'UPDPMIWR' index 5;
{$ELSE}
{$IFDEF WINDOWS}
 procedure up_write ( s: PChar );       external 'UPWINWR' index 1;
 procedure up_gotoxy ( x, y: word );    external 'UPWINWR' index 2;
 procedure up_textattr ( x: word );     external 'UPWINWR' index 3;
 procedure up_clearline;                external 'UPWINWR' index 4;
 procedure up_clrscr;  		        external 'UPWINWR' index 5;
{$ELSE}

 procedure up_write ( s: PChar );
 begin
  write ( StrPas ( s ));
 end;

 procedure up_gotoxy ( x, y: word );
 begin
  gotoxy ( x, y );
 end;

 procedure up_textattr ( x: word );
 begin
  textattr := x;
 end;

 procedure up_clearline;
 begin
  clreol;
 end;

 procedure up_clrscr;
 begin
  clrscr;
 end;
{$ENDIF}
{$ENDIF}

{ -->> For automatically unitcompiling >>-- }
function fileexist;
var f: file;
begin
 assign ( f, fn );
 {$I-}
 reset ( f );
 {$I+}
 if (ioresult = 0)or(fn='') then begin
  fileexist := true;
  close ( f );
 end else
  fileexist := false;
end;


function getdatetime ( fn: string ): longint;
var t: datetime;
    f: file;
    l: longint;
begin
 assign ( f, fn );
 reset ( f );
 getftime ( f, l );
 getdatetime := l;
 close ( f );
end;

function filenewer;
var l1, l2: longint;
begin
 l1 := getdatetime ( fn1 );
 l2 := getdatetime ( fn2 );
 filenewer := ( l1 > l2 );
end;

procedure settime ( fn1, fn2: string );
var l: longint;
    f: file;
begin
 l := getdatetime ( fn1 );
 assign ( f, fn2 );
 reset ( f );
 setftime ( f, l );
 close ( f );
end;

{
 ****************************************************************************
 Ausgaberoutinen
 ****************************************************************************
}

procedure output_chars ( s: string );
begin
 write ( outfile, s );
end;

procedure output_char ( c: char );
begin
 write ( outfile, c );
end;

procedure output (s: string);
begin
if unit_compile then
 unit_pascal_code_output ( optimize(s))
else
 writeln (outfile, optimize(s));
end;

procedure output_commentbox(s: string);
begin
output(#13#10+
       ';*********************************************************************'+#13#10+
       ';'+s+#13#10+
       ';*********************************************************************'+#13#10);
end;

procedure call_library ( libname, libfunc: string );
var lib_str: pstringstream;
    temp_s : string;
begin
  new ( lib_str, init );
  lib_str^.add_string ( libname );
  lib_str^.add_string ( libfunc );
  fargo_version_stream^.insert ( _librarycall, form_container_2, lib_str, temp_s );
  dispose ( lib_str, done );
  output ( temp_s );
end;

{
 Handles a variabledefinition.
 When it is a record, then a recursive calling is done
}

{ Function GET_VAR_SIZE -> Get the size of a variable and handle STRING
  definitions }
function get_var_size ( s: string ): word;
var sym_index : integer;
    buffer_string : string;
    m, x          : integer;
    size          : integer;
begin
if str_pos(':', s)>0 then begin
 s := trim(copy(s, pos(':', s)+1, 255));
 delete ( s, length(s), 1);
 s := trim(s);
 if pos('=', s)=0 then begin
  m := length(s);
  while (s[m]<>' ') and (m>0) do dec(m);
  s := trim(copy(s, m+1, 255));
 end;
end;
if str_pos('string', s)=0 then begin
  sym_index := data_type_index ( s );
  if sym_index <> -1 then
   get_var_size := variable_data_types[sym_index].size
  else
   get_var_size := 0;
  exit;
end else
  size := standard_string_size;
  if str_pos('[', s)>0 then begin
    buffer_string := trim(copy(s, str_pos('[', s)+1,str_pos(']', s)-str_pos('[', s)-1));
    buffer_string := calculate ( buffer_string, false );
    val(buffer_string, m, x);
    if (m<0) or (x<>0) then error(18);
    inc(m);
    size := m + m mod 2;
  end;
  if str_pos('=', s)>0 then begin
   buffer_string := copy(s, str_pos('=', s)+1, 255);
   delete ( buffer_string, str_pos(';', s), 255);
   buffer_string := trim ( buffer_string );
   buffer_string := pre_compile_string ( buffer_string );
   if str_pos('+', buffer_string) > 0 then error(18);
   size := length(buffer_string)-1;
   size := size+size mod 2;
  end;
  get_var_size := size;
end;

procedure copy_pvar_to_pvar ( var pvar_source, pvar_dest: ppascal_var; reccall: boolean );
begin
 if (pvar_source = nil) or ( pvar_source^.name = '' ) then exit;
 create_new_var ( pvar_dest );
 pvar_dest^ := pvar_source^;
 if pvar_dest^.variabelntyp = _record then
  copy_pvar_to_pvar ( pvar_dest^.record_first_var, pvar_dest^.record_first_var, true );
 if reccall then
  copy_pvar_to_pvar ( pvar_dest^.next_var, pvar_dest^.next_var, true );
end;

function handle_extended_data_types ( type_str: string; var last_var: ppascal_var; parbuild: boolean ): boolean;
var f      : file;
    ppvar  : ppascal_var;
    sn     : string;
begin
 handle_extended_data_types := false;
 if (lower_str (type_str) = 'sprites')or(lower_str ( type_str ) = 'external')or
    (lower_str (type_str) = 'file') then begin
  if type_str = 'sprites' then
   last_var^.variabelntyp := _sprites
  else
   last_var^.variabelntyp := _external;

  last_var^.const_string := copy ( last_var^.const_string, 7, 255 );
  last_var^.const_string := copy ( last_var^.const_string, 1, pos ('"', last_var^.const_string)-1 );
  if lower_str ( type_str ) = 'sprites' then
   last_var^.const_string := predirectory + sprite_dir + last_var^.const_string + '.LIB';
  if not parbuild and not not fileexist ( last_var^.const_string ) then begin
   error_txt ( 'External file not found' );
   assign ( f, last_var^.const_string );
   reset ( f, 1 );
   last_var^.size := filesize ( f ) + filesize ( f ) mod 2;
   last_var^.const_string := 'INCBIN "' + last_var^.const_string + '"';
   if filesize ( f ) mod 2 <> 0 then
    last_Var^.const_string := last_var^.const_string + #13#10 + ' dc.b 0';
   close ( f );
  end else begin
   last_var^.size := 4;
  end;
  handle_extended_data_types := true;
 end;
 { -->> Find type definition >>-- }
 ppvar := get_pascal_var_descriptor ( type_str, type_first_variable);
 if ppvar <> nil then begin
  sn := last_var^.name;
  last_var^.name := '';
  copy_pvar_to_pvar ( ppvar, last_var, false );
  last_var^.name := sn;
  handle_extended_data_types := true;
 end;
end;

function build_var_type ( var_str: string; var last_var, first_var: ppascal_var; record_call, rights,
                          parameterbuild: boolean ): longint;
var this_array_size  : word;
    record_size      : word;
    buffer_string    : string;
    buffer_string_2  : string;
    parameter_1      : string;
    parameter_2      : string;
    l1, l2           : longint;
    x                : word;
    size_count       : longint;
    total_size_count : longint;
    buffer           : ppascal_var;
    size_in_use      : word;
    sv_last          : word;
    var_used	     : boolean;
    prev_c	     : char;
begin
if copy ( var_str, 1, 4 ) = 'var ' then begin
 delete ( var_str, 1, 4 );
 var_used := true;
end else
 var_used := false;

if copy ( var_str, 1, 6 ) = 'const ' then begin
 delete ( var_str, 1, 6 );
 rights := true;
end;
var_str := trim ( var_str );
prev_c := #0;
if var_str[1] = '#' then
 if not lib_def then
  error_txt ( '"#" at the beginning is only allowed in librarydefinitions')
 else begin
  prev_c := '#';
  delete ( var_str, 1, 1 );
 end;
total_size_count := 0;
size_count := 0;
create_new_var (last_var);
last_var^.descriptor_string := '';
last_var^.const_string := '';
last_var^.variabelntyp := _no_type;
last_var^.stackpos := 0;
last_var^.var_set := false;
last_var^.prev_c := prev_c;
x := 1;
while not (var_str [ x ] in [':', '=']) do inc ( x );
if str_pos(':', var_str)>0 then begin
 buffer_string := trim(copy(var_str, 1, x-1));

if not lib_def then buffer_string := lower_str ( buffer_string );
controll_name (buffer_string);
last_var^.name := buffer_string;
buffer_string := trim (copy (var_str, str_pos(':', var_str)+1, str_pos(';', var_str) - str_pos(':', var_str)-1));
if str_pos('array', var_str) > 0 then begin
 {
   Building of a calculationstring, the compiler inserts all variables during
   compiling, when this array is needed -> %v is used for spaceidentification
 }
 buffer_string := copy (var_str, pos('[', var_str)+1, pos(']', var_str) - pos('[', var_str) - 1);
 { Reading of all parameters of this array }
 last_var^.variabelntyp := _array;
 last_var^.descriptor_string := '';
 last_var^.const_string := '';
 while buffer_string <> '' do begin
   buffer_string_2 := get_parameter (buffer_string);
   if pos('..', buffer_string_2)=0 then begin
    parameter_1 := '0';
    parameter_2 := buffer_string_2;
   end else begin
    parameter_1 := calculate(trim(copy(buffer_string_2, 1, pos('..', buffer_string_2)-1)), false);
    parameter_2 := calculate(trim(copy(buffer_string_2, pos('..', buffer_string_2)+2, 255)), false);
   end;
   val (parameter_1, l1, x);
   if x<>0 then error(13);
   val (parameter_2, l2, x);
   if x<>0 then error(13);
   { Creating of the real indexes and refresh the calculationstring }
   if last_var^.descriptor_string <> '' then
     last_var^.descriptor_string := last_var^.descriptor_string + '+' + conv_to_str (size_count)+ '*';
   if l1<0 then
    last_var^.descriptor_string := last_var^.descriptor_string + '(%v+' + conv_to_str (abs(l1))+')'
   else
    if l1>0 then
     last_var^.descriptor_string := last_var^.descriptor_string + '(%v-' + conv_to_str (l1)+')'
    else
     last_var^.descriptor_string := last_var^.descriptor_string + '%v';
   if size_count = 0 then
    inc(size_count, l2 - l1 + 1)
   else
    size_count := size_count * (l2 - l1 + 1);
   if total_size_count = 0 then
    inc(total_size_count, l2 - l1 + 1)
   else
    total_size_count := total_size_count * (l2 - l1 + 1);
 end;
end;
if str_pos('record', var_str)>0 then begin
 { Adding of a recordvariable }
 new (last_var^.record_first_var);
 last_var^.record_first_var^.name := '';
 last_var^.record_first_var^.next_var := nil;
 last_var^.record_first_var^.stackpos := 0;
 buffer := last_var^.record_first_var;
 buffer_string := get_next_line;
 size_count := 0;
 sv_last := last_variable_size;
 while buffer_string <> 'end;' do begin
  inc (size_count, build_var_type (buffer_string, buffer, buffer, true, rights, parameterbuild));
  buffer_string := get_next_line;
 end;
 last_variable_size := sv_last;
 if last_var^.descriptor_string <> '' then
  last_var^.descriptor_string := conv_to_str(size_count)+'*(' + last_var^.descriptor_string + ')';
 if total_size_count>0 then
  total_size_count := total_size_count * size_count
 else
  total_size_count := size_count;
 last_var^.size := total_size_count;
 if not rights then begin
  increase_all_variables ( last_variable_size, first_var, false);
  last_variable_size := total_size_count;
 end;
 build_var_type := total_size_count;
 if var_used then begin
  last_var^.var_t_size := last_var^.size;
  last_var^.size := 4;
  last_variable_size := 4;
 end;
 exit;
end;
{ Real datatypes handling }
end;
if (str_pos(':', var_str)=0)or(var_str[str_pos(':', var_str)+1] = ':') then begin
 { Handle an untypised variable }
 if not rights then error_txt ( 'Wrong variable definition ( no type )' );
 last_var^.name := trim ( copy ( var_str, 1, pos ('=', var_str ) - 1 ));
 controll_name ( last_var^.name );
 last_var^.variabelntyp := _no_type;
 { An '=' is necessary }
 if str_pos('=', var_str)=0 then error(15);
 buffer_string := copy(var_str, pos('=', var_str)+1, 255);
 delete (buffer_string, length(buffer_string), 1);
 buffer_string := trim(buffer_string);
 last_var^.const_string := buffer_string;
end else begin
 size_in_use := get_var_size (var_str);
 if str_pos('=', var_str)>0 then begin
  if not rights then
   error_txt ( 'Variable is a constant, but it isn''t defined as a constant' );
  if pos('''', var_str)>0 then begin
   last_var^.const_string := copy(var_str, pos('''', var_str)+1, pos(';', var_str) - pos('''',var_str) - 1);
   while last_var^.const_string[length(last_var^.const_string)] <> '''' do
    delete(last_var^.const_string, length(last_var^.const_string), 1);
   delete(last_var^.const_string, length(last_var^.const_string), 1);
   last_var^.const_string := 'dc.b "'+last_var^.const_string+'",0';
   if (length(last_var^.const_string)-8)mod 2=1 then
    last_var^.const_string := last_var^.const_string + ',0';
  end else begin
   buffer_string := trim(copy(var_str, pos('=', var_str)+1, pos(';', var_str)-pos('=',var_str)-1));
   last_var^.const_string := 'dc.l '+buffer_string;
  end;
  delete(var_str, pos('=', var_str), 255);
  var_str := var_str + ';';
 end;
 { handle typ }
 buffer_string := trim(copy(var_str, str_pos(':', var_str)+1,255));
 delete(buffer_string, length(buffer_string), 1);
 buffer_string := trim(buffer_string);
 x := length(buffer_string);
 while (buffer_string[x] <> ' ') and (x>1) do dec(x);
 buffer_string := trim(copy(buffer_String, x, 255));
 if str_pos('[', buffer_string)>0 then begin
  parameter_1 := copy (buffer_string, str_pos('[', buffer_string)+1, str_pos(']', buffer_string)-pos('[', buffer_string)-1);
  buffer_string := 'string';
  parameter_1 := calculate(parameter_1, false);
  val (parameter_1, l1, x);
  if x<>0 then error(13);
  l1 := l1 + l1 mod 2;
  last_var^.const_string := 'dc.b "';
  for x := 1 to l1 do
   last_var^.const_string := last_var^.const_string + ' ';
  last_var^.const_string := last_var^.const_string + '",0';
  if (length(last_var^.const_string)-9)mod 2=1 then
   last_var^.const_string := last_var^.const_string + ',0';
 end;
 if data_type_index (buffer_string)=-1 then begin
  if not handle_extended_data_types ( buffer_string, last_var, parameterbuild ) then
   error(16);
 end else begin
  last_var^.stackpos := 0;
  if last_var^.variabelntyp = _array then begin
   last_var^.descriptor_string := conv_to_str(variable_data_types[data_type_index(buffer_string)].size)
                                  +'*('+last_var^.descriptor_string+')';
   last_var^.arraytyp := variable_data_types[data_type_index(buffer_string)].v_type;
  end else
   last_var^.variabelntyp := variable_data_types[data_type_index(buffer_string)].v_type;
  if total_size_count > 0 then
   total_size_count := total_size_count * size_in_use
  else
   total_size_count := size_in_use;
   last_var^.size := total_size_count;
 end;
end;
 if not rights then begin
  increase_all_variables ( last_variable_size, first_var, false );
  last_variable_size := total_size_count;
 end;
 if var_used then begin
  last_var^.var_t_size := last_var^.size;
  last_var^.size := 4;
  last_variable_size := 4;
 end;
 build_var_type := total_size_count;
end;

function data_type_index (tname: string): integer;
var i: word;
begin
data_type_index := -1;
for i:=0 to high ( variable_data_types ) do
 if variable_data_types[i].name = tname then data_type_index := i;
end;

{ Reserviert Platz fr eine neue Variable }
procedure create_new_var (var last_var: ppascal_var);
var x: word;
begin
if (last_var<>nil)and(last_var^.name<>'') then begin
 new(last_var^.next_var);
 last_var^.next_var^.next_var:=nil;
 last_var^.next_var^.record_first_var:=nil;
 last_var^.next_var^.rechte := false;
 last_var^.next_var^.previous_var := last_var;
 last_var^.next_var^.var_set := false;
 last_var^.next_var^.call_level := 0;
 last_var:=last_var^.next_var;
end else begin
 if last_var=nil then
  new ( last_var );
 last_var^.next_var:=nil;
 last_var^.record_first_var:=nil;
 last_var^.rechte := false;
 last_var^.var_set := false;
 last_var^.call_level := 0;
end;
end;

{ Reserviert Platz fr eine neue Prozedur }

procedure create_new_procedure;
var x: word;
begin
if (last_procedure^.name<>'') then begin
 new (last_procedure^.next_procedure);
 last_procedure := last_procedure^.next_procedure;
 last_procedure^.next_procedure := nil;
 last_procedure^.name := '';
 last_procedure^.synonym := '';
 new (last_procedure^.first_var);
 last_procedure^.first_var^.name := '';
 last_procedure^.last_var := last_procedure^.first_var;
 last_procedure^.first_var^.record_first_var:=nil;
 last_procedure^.first_var^.next_var := nil;
 last_procedure^.asm_func := false;
end else begin
 last_procedure^.next_procedure := nil;
 if last_procedure^.first_var = nil then new (last_procedure^.first_var);
 last_procedure^.first_var^.name := '';
 last_procedure^.last_var := last_procedure^.first_var;
 last_procedure^.first_var^.record_first_var:=nil;
 last_procedure^.first_var^.next_var := nil;
 last_procedure^.synonym := '';
 last_procedure^.first_var^.next_var := nil;
 last_procedure^.asm_func := false;
end;
last_procedure^.fkt_var := nil;
end;

procedure controll_name (s: string);
var i: word;
begin
if length(s)>40 then error(7);
if s[1] in ['0'..'9','#'] then error(8);
for i:=1 to length(s) do begin
 case s[i] of
  'A'..'Z','a'..'z','0'..'9','_','#': begin end;
 else
  error(8);
 end;
end;
end;

{
 ****************************************************************************
       Einlesen aus der Quelldatei mit formatieren des Eingabestrings
 ****************************************************************************
}

{ Variabeln dieses Teiles }

{ Get a char from the sourcestream }

function getc: char;
var buffer_string: string;
begin
   if infile_string = nil then begin
    getmem ( infile_string, $200 );
    infile_str_pos := 1;
    infile_string := #0;
   end;
   if infile_str_pos >= StrLen (infile_string) then begin
     infile_str_pos := 0;
     buffer_string := '';
     infile_string := #0;
     while (length(buffer_string)=0)and not eof(infile) do begin
      readln(infile, buffer_string);
      inc(linenr);
      inc(total_line_counter);
     end;
     StrPCopy(infile_string, buffer_string);
     infile_string [ length ( buffer_string ) ] := #0;
     getc := #13;
     exit;
   end;
   inc(infile_str_pos);
   if eof(infile) and (StrLen(infile_string) = 0) then
     getc:=#0
   else
     getc:=infile_string [infile_str_pos-1];
end;

function get_next_line: string;
var output_string: string;
    c, real_c, prev_char: char;
    comment_string  : string;
    next_tk_tst: boolean;
begin
 string_read:=false;
 comment_string := '';
 output_string := '';
 last_infile_str := '';
 repeat
  c:=getc;
 until not (c in [' ',#13,#9]);
 prev_char := #0;
 if (c=#0) then begin
  output_string := '';
  exit;
 end;
 next_tk_tst := false;
 while 1=1 do begin
  real_c := c;
  case c of
' ', #13: next_tk_tst := true;
     '}': if not string_read then
           if comments=0 then warning(1) else dec(comments);
     '{': if not string_read then inc(comments);
'A'..'Z': if not string_read and (comments=0) then c:=char(ord(c)+32);
     '(': if not string_read then inc(c_rackets);
     '*': if not string_read and (prev_char in ['(','/']) then begin
           inc ( comments );
           delete ( output_string, length(output_string), 1 );
           prev_char := #0;
           c := getc;
           continue;
          end;
     '.': if not string_read and ( prev_char = '(' ) then begin
           delete ( output_string, length ( output_string ), 1);
           c := '[';
          end;
     ')': if not string_read then begin
           next_tk_tst := true;
           case prev_char of
              '*': begin
                    prev_char := #0;
                    delete ( output_string, length(output_string ), 1);
                    if comments>0 then
                     dec ( comments )
                    else begin
                     c := getc;
                     continue;
                    end;
                   end;
              '.': begin
                     delete ( output_string, length(output_string ), 1);
                     c := ']';
                   end;
           else
            if c_rackets=0 then begin
             delete ( output_string, length ( output_string ), 1 );
             warning(2);
  	    end else
 	     dec(c_rackets);
           end;
          end;
     #39: if comments=0 then string_read:=not string_read;
     '/': { DELPHI and C-styled comments '//', '*/' }
          begin
           if (prev_char = '/')and not string_read then begin
            delete (output_string, length ( output_string ) , 1 );
            delete (last_infile_str, length (last_infile_str), 1);
            prev_char := #0;
            while getc <> #13 do;
            c := getc;
            continue;
           end;
           if (prev_char = '*')and not string_read then begin
            if comments=0 then warning (1);
            prev_char := #0;
            delete (output_string, length (output_string ), 1);
            delete (last_infile_str, length (last_infile_str), 1);
            if comments>0 then
             dec ( comments );
            c := getc;
            comment_string := '';
            continue;
           end;
          end;

  end;
  if (comments>0)or(c='}')and(c<>#13) then begin
   comment_string:=comment_string+c;
   if (c='}') and (comments=0) then begin
    { Preprozessor aufrufen }
    delete (comment_string, 1, 1);
    comment_string:=trim(copy(comment_string, 1, length (comment_string)-1));
    preprozessor(comment_string);
    comment_string:='';
   end;
   prev_char := c;
   c:=getc;
   continue;
  end;
  if read_infile then begin
   if (c<>#13) then begin
    if end_char = ';' then
     output_string:=output_string+c
    else
     output_string := output_string + real_c;
    last_infile_str := last_infile_str + real_c;
   end else begin
    if (c = end_char) then break;
    if (end_char = ';')and not string_read then output_string := output_string + ' ';
   end;
  end;
  if c_rackets>0 then begin
   prev_char := c;
   c:=getc;
   continue;
  end;
  if not read_infile or string_read then begin
   prev_char := c;
   c:=getc;
   continue;
  end;
  if (c=end_char)or(c='.') then next_tk_tst := true;
  if next_tk_tst then begin
   next_tk_tst := false;
   if (end_char = ';')and(c in [#13, #9, ' ','.',';'])and(is_token (lastword(output_string, c), true, false)) then break;
   if (end_char = ':') then begin
    comment_string := lastword (output_string, c);
    if (comment_string = 'end;')or(comment_string = 'else')or(comment_string = 'of') then break;
   end;
  end;
  if (c=end_char) then break;
  prev_char := c;
  c:=getc;
 end;
 c:=getc;
 if end_char = ';' then begin
  if (c=';')or(c='.') then
   output_string:=output_string+c
  else
   if infile_str_pos > 0 then dec(infile_str_pos);
  { -->> Check for ';' in if .. then ..';'!! else >>-- }
  if (trim ( output_string ) <> 'else') then
   if (lastword(output_string, #0)='else')or(lastword(output_string, #0)='end') then begin
    insert_token ( lastword ( output_string, #0));
    delete ( output_string, length ( output_string ) - length ( lastword ( output_string, #0)), 255);
    output_string := output_string + ';';
   end;
 end else
  if infile_str_pos > 0 then dec(infile_str_pos);
 get_next_line:=trim(output_string);
end;

procedure increase_all_variables (size: word; which_var: ppascal_var; inc_last: boolean);
begin
while which_var^.next_var<>nil do begin
 inc (which_var^.stackpos, size);
 which_var := which_var^.next_var;
end;
if inc_last then inc ( which_var^.stackpos, size);
end;

procedure delete_all_variables ( var firstvar, lastvar: ppascal_var; clearall: boolean);
var buffer_pvar       : ppascal_var;
    last_buffer_pvar  : ppascal_var;
    previous          : ppascal_var;
begin
while firstvar <> nil do begin
 buffer_pvar := firstvar;
 previous := nil;
 while assigned(buffer_pvar^.next_var)and ( buffer_pvar^.name <> '') do begin
  previous := buffer_pvar;
  buffer_pvar := buffer_pvar^.next_var;
 end;
 if assigned(buffer_pvar^.record_first_var)and(buffer_pvar^.name <> '') then
  delete_all_variables ( buffer_pvar^.record_first_var, last_buffer_pvar, true );
 dispose (buffer_pvar);
 if previous <> nil then previous^.next_var := nil else firstvar := nil;
end;
if not clearall then begin
 new(firstvar);
 firstvar^.name := '';
 firstvar^.const_string := '';
 firstvar^.descriptor_string := '';
 firstvar^.next_var := nil;
 firstvar^.record_first_var := nil;
 firstvar^.stackpos := 0;
 firstvar^.size := 0;
 lastvar := firstvar;
end;
end;

{
 ****************************************************************************
  is_token
  Tests, if there is a token in the string
 ****************************************************************************
}

function is_token (x: string; lastwordtest, totallytest: boolean): boolean;
var is_i: word;
begin
is_token:=false;
for is_i := 0 to high(pascal_keywords) do
 if not totallytest then begin
  if (trim(pascal_keywords [is_i].name)=x)and(pascal_keywords [is_i].lastword=lastwordtest) then
   if pascal_keywords [is_i].keyreturn = _of then
    is_token := check_of
   else
    is_token := true;
 end else
  if (pascal_keywords[is_i].name=copy(x, 1, length(pascal_keywords[is_i].name))) then
   is_token := true;
end;

end.