
{
 -->> Objectclass for Operators >>--

 With this class you are able to create operators for floats, integers,
 strings and operator overloading

}


uses strings;

const left_side: word = $0001;
      right_side: word = $0002;

type
      operator_classes = (_int, _float, _string, _overloading );
      operators        = (_add, _sub, _mul, _idiv, _rdiv, _shl, _shr, _and,
                          _not, _or, _xor, _equal, _nonequal, _lower,
                          _lowerorequal, _bigger, _biggerorequal);

      { -->> Definition of a scanningobject >>-- }
      poperatorscanner = ^operatorscanner;
      operatorscanner = object
            constructor init;
            function scan_expression ( s: PChar; Start: PChar; direction: integer ): PChar;
            destructor done;
      end;

      poperatorclass = ^operatorclass;
      { -->> This class is for one operator, don't call this class directly -->> }
      operatorclass = object
       constructor init ( opname: string; expr_pos: word; changeable: boolean );
       procedure define_opcode ( CodeSeq: PChar ); virtual;
       { -->> Return result >>-- }
       function create_code ( CodeSeq: PChar; scannerobj: poperatorscanner ): string; virtual;
       function get_operator: PChar; virtual;
       destructor done;
      private
       operatorname         : PChar;
       CodeSequence         : PChar;
       operand_1,
       operand_2            : PChar;
       expression_position  : word;
       change_able          : boolean;
      end;

      poperatorclasslist = ^operatorclasslist;
      operatorclasslist = record
        class_pointer : poperatorclass;
        next          : poperatorclasslist;
      end;
      poperatorhandler = ^operatorhandler;
      operatorhandler = object
        constructor init;
        procedure add_operator ( opname: string;
                                 CodeSeq: PChar;
                                 childobj: poperatorclass;
                                 directions: word;
                                 operatorclass: operator_classes;
                                 changeable: boolean );
        { -->> Get valid parameter >>-- }
        function find_operator ( operator: string; operatorclass: operator_classes ): boolean;
        procedure class_scanner ( operatorclass: operator_classes; scanobject: poperatorscanner );
        destructor done;
        private
           expression_positions: array[_int .. _overloading] of word;
           first_class: array[_int .. _overloading] of poperatorclasslist;
           last_class: array[_int .. _overloading] of poperatorclasslist;
           classtype : array[_int .. _overloading] of operator_classes;
           scanners : array[_int .. _overloading] of poperatorscanner;
      end;


      { -->> All possible classes >>-- }
      pstringclass = ^stringclass;
      stringclass = object(operatorclass)
       function create_code ( CodeSeq: PChar; scanobj: poperatorscanner ): string; virtual;
      end;

      { -->> Scanner for the string-class >>-- }
      pstringscanner = ^stringscanner;
      stringscanner = object(operatorscanner)
        function scan_expression ( s: PChar; Start: PChar; direction: integer ): PChar; virtual;
      end;

      { -->> Integer class for all non float number operations >>-- }
      pintegerclass = ^integerclass;
      integerclass = object(operatorclass)
       function create_code ( CodeSeq: PChar; scanobj: poperatorscanner ): string; virtual;
      end;

      pintegerscanner = ^integerscanner;
      integerscanner = object(operatorscanner)
       function scan_expression ( s: PChar; Start: PChar; direction: integer ): PChar; virtual;
      end;

      { -->> Float class for all possible versions of real numbers >>-- }
      prealclass = ^realclass;
      realclass = object(operatorclass)
       function create_code ( CodeSeq: PChar; scanobj: poperatorscanner ): string; virtual;
      end;

      prealscanner = ^realscanner;
      realscanner = object(operatorscanner)
       function scan_expression ( s: PChar; Start: PChar; direction: integer ): PChar; virtual;
      end;

      { -->> For overloading of operators >>-- }
      poverloadclass = ^overloadclass;
      overloadclass = object(operatorclass)
       procedure define_opcode ( CodeSeq: PChar ); virtual;
       function create_code ( CodeSeq: PChar; scanobj: poperatorscanner ): string; virtual;
      end;

      poverloadscanner = ^overloadscanner;
      overloadscanner = object(operatorscanner)
        function scan_expression ( s: PChar; Start: PChar; direction: integer ): PChar; virtual;
      end;

const  operator_table : array[_add.._biggerorequal] of string[6] =
        ( '+', '-', '*', ' div ', '/', ' shl ', ' shr ', ' and ', ' not ',
          ' or ', ' xor ', '=', '<>', '<', '<=', '>', '>=' );

constructor operatorclass.init;
begin
asm
 jmp @start
 db 13,10,'void operatorclass::init()',13,10
@start:
end;
 getmem ( operatorname, length (opname) + 1 );
 StrPCopy ( operatorname, opname );
 CodeSequence := nil;
 expression_position := expr_pos;
 change_able := changeable;
end;

procedure operatorclass.define_opcode ( CodeSeq: PChar );
begin
asm
 jmp @start
 db 13,10,'void operatorclass::define_opcode(char *CodeSeq)',13,10
@start:
end;
 CodeSequence := StrNew ( CodeSeq );
end;

function operatorclass.create_code ( CodeSeq: PChar; scannerobj: poperatorscanner ): string;
var expr_1, expr_2 : PChar;
begin
asm
 jmp @start
 db 13,10,'char *operatorclass::create_code(char *CodeSeq)',13,10
@start:
end;

if expression_position and right_side<>0 then
 expr_1 := scannerobj^.scan_expression ( CodeSeq, StrPos ( CodeSeq, operatorname ), right_side )
else
 expr_1 := nil;

if expression_position and left_side<>0 then
 expr_2 := scannerobj^.scan_expression ( CodeSeq, StrPos ( CodeSeq, operatorname), left_side )
else
 expr_2 := nil;

{ -->> Insert code >>-- }

end;

function operatorclass.get_operator;
begin
asm
 jmp @start
 db 13,10,'char *operatorclass::get_operator()',13,10
@start:
end;
 get_operator := operatorname;
end;

destructor operatorclass.done;
begin
asm
 jmp @start
 db 13,10,'void operatorclass::~done()',13,10
@start:
end;
 if assigned(CodeSequence) then StrDispose ( CodeSequence );
end;


constructor operatorhandler.init;
var i: operator_classes;
begin
asm
 jmp @start
 db 13,10,'void operatorhandler::init()',13,10
@start:
end;
for i := low ( operator_classes ) to high ( operator_classes ) do begin
 scanners [ i ] := nil;
 first_class [ i ] := nil;
 last_class [ i ] := nil;
end;
end;

procedure operatorhandler.add_operator;
begin
asm
 jmp @start
 db 13,10,'void operatorhandler::add_operator()',13,10
@start:
end;
new ( childobj, init ( opname, directions, changeable ));
if first_class [ operatorclass ] = nil then begin
 new ( first_class [ operatorclass ] );
 first_class [ operatorclass ]^.class_pointer := childobj;
 last_class [ operatorclass ] := first_class [ operatorclass ];
end else begin
 new ( last_class [ operatorclass ]^.next );
 last_class [ operatorclass ]^.next^.class_pointer := childobj;
 last_class [ operatorclass ] := last_class [ operatorclass ]^.next;
end;
end;

procedure operatorhandler.class_scanner;
begin
 scanners [ operatorclass ] := scanobject;
end;

function operatorhandler.find_operator;
var class: poperatorclasslist;
    OpPCharName: PChar;
begin
asm
 jmp @start
 db 13,10,'BOOL operatorhandler::find_operator(char *operator)',13,10
@start:
end;
 getmem ( OpPCharName, length ( operator ) + 1 );
 StrPCopy ( OpPCharName, operator );
 class := first_class [ operatorclass ];
 while assigned ( class ) and
       ( StrComp (class^.class_pointer^.get_operator, OpPCharName )<>0 ) do
  class := class^.next;
 find_operator := assigned(class);
end;

destructor operatorhandler.done;
begin
asm
 jmp @start
 db 13,10,'void operatorhandler::~done()',13,10
@start:
end;
 { -->> Dispose tree >>-- }
end;


function stringclass.create_code;
begin
asm
 jmp @start_code
 db 13,10,'char *stringclass::create_code(char *CodeSeq)',13,10
@start_code:
end;
end;

{ -->> Scanner for the string-class >>-- }
function stringscanner.scan_expression;
begin
asm
 jmp @start_code
 db 13,10,'char *stringscanner::scan_expression(char *ExprStr, char *Pos, int ExprDir)',13,10
@start_code:
end;

 scan_expression := nil;
end;


function integerclass.create_code;
begin
asm
 jmp @start_code
 db 13,10,'char *integerclass::create_code()',13,10
@start_code:
end;
end;

function integerscanner.scan_expression;
begin
end;

function realclass.create_code;
begin
asm
 jmp @start_code
 db 13,10,'char *realclass::create_code()',13,10
@start_code:
end;
end;

function realscanner.scan_expression;
begin
end;

procedure overloadclass.define_opcode;
begin
asm
 jmp @start_code
 db 13,10,'void overloadclass::define_opcode(char *CodeSeq)',13,10
@start_code:
end;

end;

function overloadclass.create_code;
begin
asm
 jmp @start_code
 db 13,10,'char *overloadclass::create_code(char *CodeSeq)',13,10
@start_code:
end;
end;

function overloadscanner.scan_expression;
begin
end;

{ -->> Definition of a scanning object >>-- }
constructor operatorscanner.init;
begin
asm
 jmp @start_code
 db 13,10,'void operatorscanner::init()',13,10
@start_code:
end;

end;

function operatorscanner.scan_expression;
begin
asm
 jmp @start_code
 db 13,10,'char *operatorscanner::scan_expression(char *ExprStr, char *Pos, int ExprDir)',13,10
@start_code:
end;

 scan_expression := nil;
end;

destructor operatorscanner.done;
begin
asm
 jmp @start_code
 db 13,10,'void operatorscanner::~done()',13,10
@start_code:
end;
end;

var operator_parent_class: ^operatorhandler;
    intclass               : pintegerclass;
    floatsclass            : prealclass;
    pcharclass             : pstringclass;
    overloadedclass        : poverloadclass;

begin
new ( operator_parent_class, init );
operator_parent_class^.add_operator ( '+', 'add.%c %s,%s', intclass, left_side or right_side, _int, true );

dispose ( operator_parent_class, done );
end.