
{
  UPAS TOKEN SYSTEM

  * UltraPascal queue handling system
  * UltraPascal registerhandling
  * UltraPascal filehandling system
  * UltraPascal timetokenhandling / tokenhandling system
  * UltraPascal preprocessor
}

unit upas_tk;

interface

uses upas_def, upas_mat;

var     current_calc_regs       : array[0..15] of calc_registers;

const   { Queuezuteilung }
        begin_queue		= 6;
        con_queue		= 5;
        stp_queue		= 4;
        end_queue		= 3;
        preprozessor_queue      = 2;
        keyword_queue           = 1;
        label_queue             = 0;

{ File handling system }
procedure close_main_file;
procedure open_procedure_file;
procedure open_main_file;

{ Queue handling system }
procedure initqueues;
procedure putqueue(s: string; x: boolean; whichqueue: word);
procedure getqueue(var s: string;var x: boolean; whichqueue: word);

{ Registerhandling }
function get_freeregister: string;
function get_free_a_register: string;
procedure set_registerstate ( name: string; newstate: boolean );
function test_register ( name : string): boolean;

{ TOKEN / TIME TOKEN handlign  }
procedure insert_token ( tokenstr: string );
procedure insert_time_token ( tokenstr: string; time: word );
procedure handle_time_token;
procedure inc_time_level;
procedure dec_time_level;

{ Preprocessor }
procedure preprozessor (s: string);

implementation

uses upas_str, strings, upas_uni;

{
 Filehandling  -> Used to write procedures, ... into another file
 92+ support: The main has to be at the start!
}

procedure close_main_file;
var i: word;
begin
 if not unit_compile then begin
  {$I-}
  close ( outfile );
  {$I+}
  i := ioresult;
 end;
end;

procedure open_procedure_file;
begin
 if not unit_compile then begin
  close_main_file;
  assign ( outfile, sourcefile + '.upp' );
  append ( outfile );
 end;
end;

procedure open_main_file;
begin
 if not unit_compile then begin
  close ( outfile );
  assign ( outfile, sourcefile + '.asm' );
  append ( outfile );
 end;
end;

{
 ****************************************************************************
 Queuehandling -> bentigt, um vorherige Events zu behandeln
 ****************************************************************************
}

{       Variabeln der Queue }

type  queue_entry = record
        q_str: string[25];
        q_set: boolean;
      end;
      tqueue_entry = array[0..49] of queue_entry;
      pqueue_entry = ^tqueue_entry;

var   queues : array[0..6] of record
                queue_pointer  : pqueue_entry;
                inqueue        : word;
               end;

procedure initqueues;
var i: word;
begin
for i:=0 to high(queues) do begin
 if queues[i].queue_pointer = nil then
  new(queues[i].queue_pointer);
 queues[i].inqueue:=0;
end;
end;

procedure putqueue(s: string; x: boolean; whichqueue: word);
begin
 inc(queues[whichqueue].inqueue);
 queues[whichqueue].queue_pointer^[queues[whichqueue].inqueue].q_str:=s;
 queues[whichqueue].queue_pointer^[queues[whichqueue].inqueue].q_set:=x;
end;

procedure getqueue(var s: string;var x: boolean; whichqueue: word);
begin
 if queues[whichqueue].inqueue=0 then
  s:=''
 else begin
  s:=queues[whichqueue].queue_pointer^[queues[whichqueue].inqueue].q_str;
  x:=queues[whichqueue].queue_pointer^[queues[whichqueue].inqueue].q_set;
  dec(queues[whichqueue].inqueue);
 end;
end;

{
 ****************************************************************************
 Verwaltung der Register um unerwnschte Ergebniss zu vermeiden
 ****************************************************************************
}

function get_freeregister: string;
var i: word;
begin
for i:=0 to high(current_calc_regs) do
 if not current_calc_regs[i].in_use then begin
  current_calc_regs[i].in_use := true;
  get_freeregister := current_calc_regs[i].name;
  exit;
 end;
end;

{ -->> Register optimization >>-- }
function search_register_contents (search_for: string): string;
begin
end;

procedure set_register_content (regname, regstate: string);
begin
end;


function get_free_a_register: string;
var i: word;
begin
for i := 8 to 8+4 do
 if not current_calc_regs[i].in_use then begin
  current_calc_regs[i].in_use := true;
  get_free_a_register := current_calc_regs[i].name;
  exit;
 end;
end;

procedure set_registerstate ( name: string; newstate: boolean );
var i: word;
begin
for i:=0 to high(current_calc_regs) do
 if current_calc_regs[i].name = name then
  current_calc_regs[i].in_use := newstate;
end;

function test_register ( name : string): boolean;
var i: word;
begin
 if str_pos('(', name)>0 then name := copy ( name, str_pos('(', name)+1, 2 );
 test_register := false;
 for i:=0 to high(current_calc_regs) do
  if current_calc_regs[i].name = name then test_register:=true;
end;

{
 ****************************************************************************
 Token-handling

 Token handling is to handle 'begin' ... 'end', if they aren't founded in
 an 'if', 'for', ..., whatever. With this it is possible to insert a line
 "virtually" into the source, so no special handling from the procedures is
 necessary. The 'Token-handler' is always after one instruction called. One
 Instruction are marked in the token list, so the compiler knows it.
 The 'Token-handler' is called from the parser befor it exits
 ****************************************************************************
}

type ptoken_list = ^token_list;
     token_list	= array[0..49] of record
        s: string[80];
        time: word;
        level_time: word;	{ EX: for x := 0 to 1 do
	                               repeat  // INC LEVEL

                                       until   // DEC LEVEL
                                      // OUT OF FOR
                                 }
     end;

var time_list	: ptoken_list;

procedure insert_token ( tokenstr: string );
var p1, p2 : pchar;
begin
 tokenstr := tokenstr + #13;
 getmem ( p1, $200 );
 getmem ( p2, $200 );
 StrCopy ( p2, infile_string + infile_str_pos );
 StrPCopy ( p1, tokenstr );
 StrCat ( p1, p2 );
 StrCopy ( infile_string, p1 );
 freemem ( p2, $200);
 freemem ( p1, $200);
 infile_str_pos := 0;
end;

procedure insert_time_token ( tokenstr: string; time: word );
var i: word;
begin
 if time_list = nil then begin
  new ( time_list );
  for i := 0 to high ( token_list ) do time_list^[i].s := '';
 end;
 for i := 0 to high( token_list ) do
  if time_list^[i].s = '' then begin
   time_list^[i].s := tokenstr;
   time_list^[i].time := time;
   time_list^[i].level_time := 0;
   exit;
  end;
end;

procedure handle_time_token;
var i : word;
begin
for i := 0 to high ( token_list ) do begin
 if time_list^[i].level_time = 0 then
  dec ( time_list^[i].time );
 if (time_list^[i].s <> '')and(time_list^[i].time = 0) then begin
  insert_token ( time_list^[i].s );
  time_list^[i].s := '';
 end;
end;
end;

procedure inc_time_level;
var i: word;
begin
for i := 0 to high ( token_list ) do
 if time_list^[i].time > 0 then inc ( time_list^[i].level_time );
end;

procedure dec_time_level;
var i: word;
begin
for i := 0 to high ( token_list ) do
 if (time_list^[i].level_time > 0) then dec ( time_list^[i].level_time );
handle_time_token;
end;

function getindex (s: string): integer;
var i: word;
begin
for i:=0 to 99 do
 if defined_names^[i].name=s then break;
if i=99 then getindex:=-1 else getindex:=i;
end;

procedure preprozessor (s: string);
var prestring: string;
    sym_index: integer;
    queue_string: string;
    queue_set: boolean;
    prev_s : string;
begin
 s:=trim(s);
 if s[1]<>'$' then exit;
 if unit_compile then begin
  open_preprocessor_code ( s );
  exit;
 end;
 delete(s,1,1);
 if pos(' ',s)=0 then prestring:=s else prestring:=copy(s,1,pos(' ',s)-1);
 delete(s,1,length(prestring)+1);
 prev_s := s;
 s:=upper_str(trim(s));
 if prestring = 'IFDEF' then begin
  { Definiertes Symbol suchen }
  if not read_infile then begin
   { To allow 'IF' in 'IF'  -> stackprincip }
   putqueue('if', false, preprozessor_queue);
   putqueue('if', true, preprozessor_queue);
   exit;
  end;
  putqueue ( 'if', read_infile, preprozessor_queue );
  sym_index:=getindex(s);
  if (sym_index=-1) then
   read_infile:=false
  else
   if defined_names^[sym_index].defined then
    read_infile:=true
   else
    read_infile:=false;
  putqueue('if', read_infile, preprozessor_queue);
 end;
 if prestring = 'IFNDEF' then begin
  if not read_infile then begin
   { To allow 'IF' in 'IF'  -> stackprincip }
   putqueue ('if', false, preprozessor_queue );
   putqueue ('if', true, preprozessor_queue);
   exit;
  end;
  putqueue('if', read_infile, preprozessor_queue);
  sym_index:=getindex(s);
  if (sym_index=-1) then
   read_infile:=true
  else
   if not defined_names^[sym_index].defined then
    read_infile:=true
   else
    read_infile:=false;
  putqueue('if', read_infile, preprozessor_queue);
 end;
 if prestring = 'ELSE' then begin
  { 'ELSE' Teil einer IF(N)DEF Konstruktion }
  getqueue(queue_string, queue_set, preprozessor_queue);
  if queue_string<>'if' then error(4);
  read_infile:=not queue_set;
  putqueue('if', read_infile, preprozessor_queue);
 end;
 if prestring = 'ENDIF' then begin
  { 'ENDIF' einer IF(N)DEF Konstruktion }
  getqueue(queue_string, queue_set, preprozessor_queue);
  getqueue(queue_string, queue_set, preprozessor_queue);
  if queue_string<>'if' then error(4);
  read_infile:=queue_set;
 end;

 { ---<<  Here was a bug, preprozessorinstructions are catched, although it
          shouldn't be done <<--- }
 if not read_infile then exit;

 if prestring = 'DEFINE' then begin
  sym_index:=getindex(s);
  if (sym_index=-1) then sym_index:=getindex('');
  if (sym_index=-1) then error(5);
  defined_names^[sym_index].name:=s;
  defined_names^[sym_index].defined:=true;
 end;
 if prestring = 'UNDEF' then begin
  sym_index:=getindex(s);
  if (sym_index <> -1) then defined_names^[sym_index].defined:=false;
 end;
 if prestring = 'S' then begin
  s := calculate ( s, false);
  val (s, program_stack_size, sym_index);
  if sym_index <> 0 then error(14);
 end;
 if prestring = 'STRING' then begin
  s := calculate ( s, false );
  val (s, standard_string_size, sym_index);
  if sym_index <> 0 then error(19);
 end;
 if prestring = 'MESSAGE' then begin
  up_gotoxy ( 1, 19 );
  up_con_write ( prev_s );
 end;
 if prestring = 'ERROR' then begin
  up_textattr($C);
  up_gotoxy(1,18);
  up_con_write('Errormessage from source: '#13#10+prev_s);
  halt;
 end;
end;

end.