
/*****************************************************************************\
*     UltraPascal SYSTEM Unit
*------------------------------------------------------------------------------
*
*    This unit defines all availabel procedures / synonyms, to support
*    different versions of:
*         TI92 assemblylanguages (Fargo I, Fargo II, TIE, 92+)
*         TI89 assemblylanguages (PlusShell, DoorsOS)
*	  PFA support (PlusShell for all)
*
*    Don't modify this unit, MXM Projects will update it, send us
*    suggestions.
*
*    created by           MXM-TriX        mxmtrix@gmx.net
*    released by          MXM Projects    http://start.at/mxm
*
*    Last update	  06091998
\*****************************************************************************/


unit 'system';

interface

const 
/*
  Here are untyped constants used, because they don't use any code, they are
  inserted into the source
*/
// "Colors"
        normal_color = 4;
        invers_color = 0;
 
// Font's
        little_font  = 0;        
        medium_font  = 1;
        big_font     = 2;

// boolean
	false        = 0;
        true         = -1;      // Scc instructions use -1 not 1!
{$IFDEF __FARGO_1}
        lcd_mem      = $4440;   // ROM 1.X version
{$ELSE}
// Here something like a synonym is used
        lcd_mem      = tios::main_lcd;
{$ENDIF}

/* 
  Here the system unit defines the various types, sizes between TI92 & TI89
  ( screensize for example )
*/
		  
{$IFNDEF __89_CODE}
	max_x	     = 240;
	max_y	     = 128;
{$ELSE}
	max_x        = 160;
	max_y        = 100;
{$ENDIF}

var
// Text output variables
      x_location, y_location, textattr: word;


function readkey : word;		// Support for the keyboard

// Screen / Pixel routines
procedure clrscr;			// Show stateline
procedure zapscr;       		// Don't show stateline
procedure put_pixel ( x, y: word );	// Set pixel x, y
procedure clear_pixel ( x, y: word);	// Clear pixel x, y

// String output routines

procedure write ( str: string );	// Write s to x_location, y_location
procedure setfont (nr: word);		// Set outputfont
procedure gotoxy ( x, y: word );	// Set x_location, y_location to x, y

// String handling routines

function length ( var s: string ): word;	// String length

// Float handling routines

procedure ftoa ( var s: string; r: real );	// F(loat)TO A(scii)

// Memory handling routines

function getmem ( size: longint ): pointer;	// Get a memoryblock (PasLib)
procedure freemem ( ptr: pointer);		// Free a memoryblock (PasLib)

implementation

/******************************************************************************
 SYSTEM unit for
 FARGO I
******************************************************************************/

{$IFDEF __FARGO_1}

 // Compiling info
 {$MESSAGE Compiling for Fargo I}

uses flib,romlib;

procedure clrscr;
synonym 'flib::clr_scr()';
procedure zapscr;
synonym 'flib::clr_scr()';

function readkey : word;
synonym 'flib::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'flib::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'flib::pixel_off(x, y)';

procedure write (str: string);
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 // Shorter code too!
 romlib::puttext ( x_location, y_location, str, textattr );
end;

procedure setfont(nr: word);
synonym 'romlib::set_font(nr)';

function length;
synonym 'romlib::strlen(s)';           // Is that right?

// Isn't supported in Fargo I
procedure ftoa; assembler;
asm
end;

{$ENDIF}

{$IFDEF __FARGO_2}
/******************************************************************************
 SYSTEM unit for
 FARGO II 
 ******************************************************************************/

 // Informations about compiling
 {$MESSAGE Compiling for Fargo II}

uses flib,tios; // These routines are automatically included

procedure clrscr;
synonym 'flib::clr_scr()';

procedure zapscr;
synonym 'flib::zap_screen()';

function readkey: word;
synonym 'flib::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'flib::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'flib::pixel_off(x, y)';

procedure write ( str: string );
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 // Secondly it saves the size of the code a little bit ( maybe )
 tios::DrawStrXY(x_location, y_location, str, textattr);
end;

procedure setfont (nr: word);
synonym 'tios::FontSetSys(nr)';

function length;
synonym 'tios::strlen(s)';

/*
  FTOA by Rafael H. Padilla   ( dysup@mcsa.net.mx )
  It is a completly assemblyroutine, so assembler is used ( although it
  generates the same entry code )
*/

procedure ftoa; assembler;
asm
   move.l 4(a5),a1
   move.l 0(a5),d7
   move.l d7,d0
   jsr ffp68343::ffpfpa
   lea 14(sp),a3
   move.l sp,a4
   move.w #13,d1
next:
   move.b (a4)+,(a1)+
   dbf.w d1,next
   move.l a3,sp
end;

{$ENDIF}

{$IFDEF __TIE}
/******************************************************************************
 SYSTEM unit for
 TIE
 ******************************************************************************/

 // Informations about compiling
 {$MESSAGE Compiling for TIE}

uses flib,tios; // These routines are automatically included

procedure clrscr;
synonym 'flib::clr_scr()';

procedure zapscr;
synonym 'flib::zap_screen()';

function readkey: word;
synonym 'flib::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'flib::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'flib::pixel_off(x, y)';

procedure write ( str: string );
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 // Secondly it saves the size of the code a little bit ( maybe )
 tios::DrawStrXY(x_location, y_location, str, textattr);
end;

procedure setfont (nr: word);
synonym 'tios::FontSetSys(nr)';

function length;
synonym 'tios::strlen(s)';

/*
  FTOA by Rafael H. Padilla   ( dysup@mcsa.net.mx )
  It is a completly assemblyroutine, so assembler is used ( although it
  generates the same entry code )
*/

procedure ftoa; assembler;
asm
   move.l 4(a5),a1
   move.l 0(a5),d7
   move.l d7,d0
   jsr ffp68343::ffpfpa
   lea 14(sp),a3
   move.l sp,a4
   move.w #13,d1
next:
   move.b (a4)+,(a1)+
   dbf.w d1,next
   move.l a3,sp
end;

{$ENDIF}

{$IFDEF __92_PLUS}

/*****************************************************************************\
 SYSTEM unit for
  TI92+ support
\*****************************************************************************/
 {$MESSAGE creating a TI89 program}
uses tios,util;

function readkey: word;
synonym 'util::idle_loop';

procedure clrscr;
synonym 'util::clr_scr()';

procedure zapscr;
synonym 'util::zap_screen()';

function readkey: word;
synonym 'util::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'util::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'util::pixel_off(x, y)';

procedure write ( str: string );
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 util::DrawStrXY(x_location, y_location, str, textattr);
end;

procedure setfont (nr: word);
synonym 'tios::FontSetSys(nr)';

function length;
synonym 'tios::strlen(s)';


procedure ftoa; assembler;
asm
   move.l 4(a5),a1
   move.l 0(a5),d7
   move.l d7,d0
   jsr ffp68343::ffpfpa
   lea 14(sp),a3
   move.l sp,a4
   move.w #13,d1
next:
   move.b (a4)+,(a1)+
   dbf.w d1,next
   move.l a3,sp
end;


{$ENDIF}

{$IFDEF __89_CODE}

/*****************************************************************************\
 SYSTEM unit for
  TI89 support
\*****************************************************************************/
 {$MESSAGE creating a TI89 program}
uses tios,util;

function readkey: word;
synonym 'util::idle_loop';

procedure clrscr;
synonym 'util::clr_scr()';

procedure zapscr;
synonym 'util::zap_screen()';

function readkey: word;
synonym 'util::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'util::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'util::pixel_off(x, y)';

procedure write ( str: string );
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 util::DrawStrXY(x_location, y_location, str, textattr);
end;

procedure setfont (nr: word);
synonym 'tios::FontSetSys(nr)';

function length;
synonym 'tios::strlen(s)';


procedure ftoa; assembler;
asm
   move.l 4(a5),a1
   move.l 0(a5),d7
   move.l d7,d0
   jsr ffp68343::ffpfpa
   lea 14(sp),a3
   move.l sp,a4
   move.w #13,d1
next:
   move.b (a4)+,(a1)+
   dbf.w d1,next
   move.l a3,sp
end;

{$ENDIF}

/******************************************************************************
 SYSTEM unit for
   PFA (PlusShell for all)
******************************************************************************/

{$IFDEF __PFA}
/*****************************************************************************\
 SYSTEM unit for
  PFA support (TI92, TI92+, TI89)
\*****************************************************************************/
 {$MESSAGE creating a TI89, TI92+ and TIE program}
uses tios,util;

function readkey: word;
synonym 'util::idle_loop';

procedure clrscr;
synonym 'util::clr_scr()';

procedure zapscr;
synonym 'util::zap_screen()';

function readkey: word;
synonym 'util::idle_loop()';

procedure put_pixel ( x, y: word );
synonym 'util::pixel_on(x, y)';

procedure clear_pixel ( x, y: word);
synonym 'util::pixel_off(x, y)';

procedure write ( str: string );
begin
 // Not possible in a synonym, because str can be a stringmanipulation!
 util::DrawStrXY(x_location, y_location, str, textattr);
end;

procedure setfont (nr: word);
synonym 'tios::FontSetSys(nr)';

function length;
synonym 'tios::strlen(s)';


procedure ftoa; assembler;
asm
   move.l 4(a5),a1
   move.l 0(a5),d7
   move.l d7,d0
   jsr ffp68343::ffpfpa
   lea 14(sp),a3
   move.l sp,a4
   move.w #13,d1
next:
   move.b (a4)+,(a1)+
   dbf.w d1,next
   move.l a3,sp
end;

{$ENDIF}

// Routines, which are availabel in all versions
// PASLIB isn't converted to Fargo I!!!

function getmem;
synonym 'paslib::getmem(size)';

procedure freemem;
synonym 'paslib::freemem(ptr)';

procedure gotoxy;
begin
 x_location := x;
 y_location := y;
end;

// SYSTEM entrycode ( always imported to an assemblyprogram )
begin
 x_location := 1;
 y_location := 1;
 textattr := 4;
end.                    